from mountaineer.client_compiler.postcss import PostCSSBundler
from mountaineer.plugin import BuildConfig, MountaineerPlugin

from mountaineer_exceptions.controllers.exception_controller import ExceptionController
from mountaineer_exceptions.views import get_core_view_path

plugin = MountaineerPlugin(
    name="mountaineer-exceptions",
    controllers=[ExceptionController],
    view_root=get_core_view_path(""),
    build_config=BuildConfig(custom_builders=[PostCSSBundler()]),
)



import inspect
import linecache
import traceback
from pathlib import Path
from uuid import UUID, uuid4

from pydantic import BaseModel
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import PythonLexer, guess_lexer_for_filename
from pygments.util import ClassNotFound


class ExceptionFrame(BaseModel):
    id: UUID
    file_name: str
    line_number: int
    function_name: str
    local_values: dict[str, str]
    code_context: str
    start_line_number: int
    end_line_number: int


class ParsedException(BaseModel):
    exc_type: str
    exc_value: str
    frames: list[ExceptionFrame]


class ExceptionParser:
    def __init__(self):
        self.formatter = HtmlFormatter(style="github-dark")
        self.python_lexer = PythonLexer()

    def _get_lexer(self, filename: str, code: str):
        try:
            return guess_lexer_for_filename(filename, code)
        except ClassNotFound:
            return self.python_lexer

    def _get_context(
        self, filename: str, lineno: int, context_lines: int = 5
    ) -> tuple[str, int, int]:
        """
        Get the code context and starting line number for the given error location.

        :param filename: Path to the source file
        :param lineno: Line number where the error occurred
        :param context_lines: Number of lines to show before and after the error

        """
        start_line = max(lineno - context_lines, 1)  # Don't go below line 1
        end_line = lineno + context_lines + 1

        lines = []
        for i in range(start_line, end_line):
            line = linecache.getline(filename, i)
            if line:
                lines.append(line)
        code = "".join(lines)

        lexer = self._get_lexer(filename, code)
        highlighted = highlight(code, lexer, self.formatter)

        return highlighted, start_line, end_line

    def _format_value(self, value: object) -> str:
        try:
            if inspect.isclass(value) or inspect.isfunction(value):
                return str(value)
            formatted = highlight(repr(value), self.python_lexer, self.formatter)
            return formatted
        except Exception:
            return str(value)

    def parse_exception(self, exc: BaseException) -> ParsedException:
        frames = []
        tb = traceback.extract_tb(exc.__traceback__)

        for frame_summary in tb:
            filename = frame_summary.filename
            lineno = frame_summary.lineno
            function = frame_summary.name

            # Get locals from the frame
            frame = None
            tb_frame = exc.__traceback__
            while tb_frame is not None:
                if (
                    tb_frame.tb_frame.f_code.co_filename == filename
                    and tb_frame.tb_lineno == lineno
                ):
                    frame = tb_frame.tb_frame
                    break
                tb_frame = tb_frame.tb_next

            locals_dict = {}
            if frame is not None:
                for key, value in frame.f_locals.items():
                    if not key.startswith("__"):
                        locals_dict[key] = self._format_value(value)

            code_context, start_line, end_line = self._get_context(
                filename, lineno or -1
            )

            frames.append(
                ExceptionFrame(
                    id=uuid4(),
                    file_name=self.get_package_path(filename),
                    line_number=lineno or -1,
                    function_name=function,
                    code_context=code_context,
                    local_values=locals_dict,
                    start_line_number=start_line,
                    end_line_number=end_line,
                )
            )

        return ParsedException(
            exc_type=exc.__class__.__name__, exc_value=str(exc), frames=frames
        )

    def get_style_defs(self) -> str:
        """Get CSS style definitions for syntax highlighting"""
        return self.formatter.get_style_defs()  # type: ignore

    def get_package_path(self, filepath: str) -> str:
        """
        Extract the relevant package path from a full system path.

        Args:
            filepath: Full system path to a Python file

        Returns:
            Shortened path relative to closest parent package
        """
        path = Path(filepath)

        # Find closest parent directory with __init__.py
        current = path.parent
        package_root = None

        while True:
            if (current / "__init__.py").exists():
                package_root = current
                current = current.parent
            else:
                break

        if package_root is None:
            # No package found, use filename only
            return path.name

        # Get relative path from package root
        try:
            rel_path = path.relative_to(package_root.parent)
            return str(rel_path)
        except ValueError:
            # Fallback to filename if relative_to fails
            return path.name



from mountaineer.cli import handle_build

from mountaineer_exceptions.plugin import plugin

app = plugin.to_webserver()


def build():
    handle_build(webcontroller="mountaineer_exceptions.cli:app")



from mountaineer.controller import ControllerBase
from mountaineer.render import LinkAttribute, Metadata, RenderBase

from mountaineer_exceptions.controllers.traceback import (
    ExceptionParser,
    ParsedException,
)


class ExceptionRender(RenderBase):
    exception: str
    stack: str | None

    # CSS that should be injected into the page to properly view the exceptions
    formatting_style: str
    parsed_exception: ParsedException


class ExceptionController(ControllerBase):
    """
    Controller intended for internal use only. Allows our development server
    to render exceptions in the browser and leverage our SSR-injected live reloading.

    """

    url = "/_exception"
    view_path = "core/exception/page.tsx"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.traceback_parser = ExceptionParser()

    def render(
        self, exception: str, stack: str, parsed_exception: ParsedException
    ) -> ExceptionRender:
        # Reverse the frames so they show with more specific errors at the top of the page
        parsed_exception.frames.reverse()

        # Exceptions can't be passed through as API types, so parents should pre-convert them to
        # strings before rendering
        return ExceptionRender(
            exception=exception,
            stack=stack,
            formatting_style=self.traceback_parser.get_style_defs(),
            parsed_exception=parsed_exception,
            metadata=Metadata(
                title=f"Exception: {exception}",
                links=[
                    LinkAttribute(
                        rel="stylesheet", href=f"{self._scripts_prefix}/core_main.css"
                    )
                ],
                ignore_global_metadata=True,
            ),
        )



import importlib.util
from pathlib import Path
from textwrap import dedent

import pytest

from mountaineer_exceptions.controllers.traceback import (
    ExceptionParser,
    ParsedException,
)


def nested_function(x: int) -> None:
    local_var = "test string"  # noqa: F841
    nested_dict = {"key": "value"}  # noqa: F841
    raise ValueError(f"Test error with {x}")


def function_with_locals() -> None:
    x = 42
    y = "string"  # noqa: F841
    z = [1, 2, 3]  # noqa: F841
    nested_function(x)


@pytest.fixture
def parser():
    return ExceptionParser()


@pytest.fixture
def complex_exception() -> ValueError:
    try:
        function_with_locals()
    except ValueError as e:
        return e
    raise RuntimeError("Expected ValueError was not raised")


def test_basic_exception_parsing(parser):
    try:
        raise ValueError("Test error")
    except ValueError as e:
        result = parser.parse_exception(e)

    assert result.exc_type == "ValueError"
    assert result.exc_value == "Test error"
    assert len(result.frames) > 0

    frame = result.frames[-1]
    assert frame.file_name.endswith("test_traceback.py")
    assert isinstance(frame.line_number, int)
    assert frame.function_name == "test_basic_exception_parsing"
    assert isinstance(frame.code_context, str)
    assert "<span" in frame.code_context
    assert isinstance(frame.local_values, dict)


def test_nested_exception_with_locals(parser, complex_exception):
    result = parser.parse_exception(complex_exception)

    assert result.exc_type == "ValueError"
    assert "Test error with 42" in result.exc_value
    assert len(result.frames) >= 2

    nested_frame = None
    for frame in result.frames:
        if frame.function_name == "nested_function":
            nested_frame = frame
            break

    assert nested_frame is not None
    assert "local_var" in nested_frame.local_values
    assert "nested_dict" in nested_frame.local_values
    assert isinstance(nested_frame.local_values["local_var"], str)
    assert "<span" in nested_frame.local_values["nested_dict"]


def test_syntax_highlighting(parser):
    try:
        x = {"complex": [1, 2, 3], "dict": {"nested": True}}  # noqa: F841
        raise Exception("Test")
    except Exception as e:
        result = parser.parse_exception(e)

    frame = result.frames[-1]
    assert "class=" in frame.code_context
    assert "span" in frame.code_context
    assert frame.local_values["x"].count("<span") > 1


def test_different_file_types(parser, tmp_path):
    py_file = tmp_path / "test.py"
    py_file.write_text(
        dedent(
            """
        def test_function():
            x = 42
            raise ValueError("Test")
    """
        )
    )

    spec = importlib.util.spec_from_file_location("test_module", py_file)
    assert spec

    module = importlib.util.module_from_spec(spec)
    assert module
    assert spec.loader

    spec.loader.exec_module(module)

    result: ParsedException | None = None
    try:
        module.test_function()
    except ValueError as e:
        result = parser.parse_exception(e)

    assert result

    frame = result.frames[-1]
    assert frame.file_name == "test.py"
    assert "test_function" in frame.code_context
    assert "<span" in frame.code_context


def test_get_style_defs(parser):
    styles = parser.get_style_defs()
    assert isinstance(styles, str)
    assert len(styles) > 0


def test_exception_without_traceback(parser):
    exc = ValueError("Test error")
    result = parser.parse_exception(exc)

    assert result.exc_type == "ValueError"
    assert result.exc_value == "Test error"
    assert isinstance(result.frames, list)


def test_line_numbers_accuracy(parser):
    def error_on_specific_line():
        x = 1  # noqa: F841
        y = 2  # noqa: F841
        raise ValueError("Test")

    result: ParsedException | None = None
    try:
        error_on_specific_line()
    except ValueError as e:
        result = parser.parse_exception(e)

    assert result
    frame = result.frames[-1]
    assert "ValueError" in frame.code_context
    assert frame.line_number > 0


@pytest.fixture
def mock_project(tmp_path):
    """Create a mock project structure with packages"""
    # Main package
    package_root = tmp_path / "myproject"
    package_root.mkdir()
    (package_root / "__init__.py").touch()

    # Subpackage
    subpackage = package_root / "subpackage"
    subpackage.mkdir()
    (subpackage / "__init__.py").touch()

    # Some modules
    (package_root / "module.py").touch()
    (subpackage / "submodule.py").touch()

    # Non-package directory
    outside = tmp_path / "outside"
    outside.mkdir()
    (outside / "standalone.py").touch()

    return tmp_path


@pytest.mark.parametrize(
    "project_path, expected_path",
    [
        ("myproject/module.py", "myproject/module.py"),
        ("myproject/subpackage/submodule.py", "myproject/subpackage/submodule.py"),
        ("outside/standalone.py", "standalone.py"),
    ],
)
def test_package_module(
    project_path: str, expected_path: str, mock_project: Path, parser: ExceptionParser
):
    path = str(mock_project / project_path)
    assert parser.get_package_path(path) == expected_path


def test_nonexistent_path(parser: ExceptionParser):
    path = "/nonexistent/path/file.py"
    assert parser.get_package_path(path) == "file.py"


def test_nested_package_structure(tmp_path, parser: ExceptionParser):
    # Create deep nested structure
    structure = tmp_path / "root" / "pkg1" / "pkg2" / "pkg3"
    structure.mkdir(parents=True)

    # Create __init__.py files at different levels
    (tmp_path / "root" / "pkg1" / "__init__.py").touch()
    (tmp_path / "root" / "pkg1" / "pkg2" / "__init__.py").touch()
    (tmp_path / "root" / "pkg1" / "pkg2" / "pkg3" / "__init__.py").touch()

    test_file = structure / "module.py"
    test_file.touch()

    assert parser.get_package_path(str(test_file)) == "pkg1/pkg2/pkg3/module.py"
